#include "defines.h"
#include "memory_compression.h"

#if defined(NEC_MEMORY_COMPRESSION)
#include <limits.h>
#include <memory.h>
#include <assert.h>

#ifndef max
#define max(a,b) ((a) > (b) ? (a) : (b))
#endif
#ifndef min
#define min(a,b) ((a) < (b) ? (a) : (b))
#endif

static unsigned int QuantizePixel(int org, int pred, int N, unsigned int* q_table, int table_size) 
{
  int delta = org - pred;
  unsigned int mod = (delta + N) % N;
  int tbl_idx;
  unsigned int rec[2];

  for(tbl_idx = 0; mod>q_table[tbl_idx] && tbl_idx<table_size; tbl_idx++);

  rec[0] = (pred + q_table[tbl_idx % table_size] + N) % N;
  rec[1] = (pred + q_table[(tbl_idx-1 + table_size) % table_size] + N) % N;
  return rec[absm(org-(int)rec[1]) < absm(org-(int)rec[0]) ? 1 : 0];
}

void MemoryCompressionProcess(ImageParameters *img, imgpel **imgY, imgpel ***imgUV)
{
  int i, j, k, uv;
  int interval, minterval;
  unsigned int N, table_size, *q_table;
  int width, height;
  unsigned int pred;

  //luminance
  width = img->width;
  height = img->height;
  interval = active_sps->dpcm_luma_interval_minus1 + 1;
  N = 1 << (active_sps->bit_depth_luma_minus8+8);
  table_size = 1 << (active_sps->dpcm_luma_bit_depth_minus1+1);
  q_table = &active_sps->dpcm_luma_qmatrix[0];
  for (j=0; j<height; j++) {
    for (i=0; i<width; i+=interval) {
      minterval = ((i+interval)>width) ? width-i : interval;
      pred = imgY[j][i];
      for (k=1; k<minterval; k++) {
        imgY[j][i+k] = pred = QuantizePixel(imgY[j][i+k], pred, N, q_table, table_size);
      }
    }
  }

  if (!imgUV) {
    return;
  }
  //chrominance
  width = img->width_cr;
  height = img->height_cr;
  interval = active_sps->dpcm_chroma_interval_minus1 + 1;
  N = 1 << (active_sps->bit_depth_chroma_minus8+8);
  table_size = 1 << (active_sps->dpcm_chroma_bit_depth_minus1+1);
  q_table = &active_sps->dpcm_chroma_qmatrix[0];
  for (uv=0; uv<2; uv++) {
    for (j=0; j<height; j++) {
      for (i=0; i<width; i+=interval) {
        minterval = ((i+interval)>width) ? width-i : interval;
        pred = imgUV[uv][j][i];
        for (k=1; k<minterval; k++) {
          imgUV[uv][j][i+k] = pred = QuantizePixel(imgUV[uv][j][i+k], pred, N, q_table, table_size);
        }
      }
    }
  }
}


void SetDefaultDPCMQMatrix(unsigned int* dst_qmatrix, int bit_depth, int dpcm_bit_depth)
{
  static const unsigned int default_dpcm_qmatrix_tab_8to5[32] =
  {
      0,  1,  2,  3,
      6,  9, 14, 18,
     24, 31, 39, 48,
     58, 72, 86,103,
    128,153,170,184,
    198,208,217,225,
    232,238,242,247,
    250,253,254,255,
  };
  static const unsigned int default_dpcm_qmatrix_tab_12to7[128] =
  {
       0,   4,   8,  12,  16,  20,  24,  28,  32,  36,  40,  44,  48,  60,  72,  84,
      96, 108, 120, 132, 144, 164, 184, 204, 224, 240, 256, 272, 288, 312, 336, 360,
     384, 412, 440, 468, 496, 528, 560, 592, 624, 660, 696, 732, 768, 808, 848, 888,
     928, 984,1040,1096,1152,1208,1264,1320,1376,1444,1512,1580,1648,1748,1848,1948,
    2048,2148,2248,2348,2448,2516,2584,2652,2720,2776,2832,2888,2944,3000,3056,3112,
    3168,3208,3248,3288,3328,3364,3400,3436,3472,3504,3536,3568,3600,3628,3656,3684,
    3712,3736,3760,3784,3808,3824,3840,3856,3872,3892,3912,3932,3952,3964,3976,3988,
    4000,4012,4024,4036,4048,4052,4056,4060,4064,4068,4072,4076,4080,4084,4088,4092,
  };

  int i;
  const unsigned int *src_qmatrix;

  assert((bit_depth == 8 && dpcm_bit_depth == 5) || (bit_depth == 12 && dpcm_bit_depth == 7));

  if (bit_depth == 8 && dpcm_bit_depth == 5) {
    src_qmatrix = &default_dpcm_qmatrix_tab_8to5[0];
  } else if (bit_depth == 12 && dpcm_bit_depth == 7) {
    src_qmatrix = &default_dpcm_qmatrix_tab_12to7[0];
  }

  for (i = 0; i < 1<<dpcm_bit_depth; i++) {
    dst_qmatrix[i] = src_qmatrix[i];
  }
}


Boolean ParseDPCMQMatrixFile(char* filename, unsigned int* dst_qmatrix, int qmatrix_size)
{
  int idx;
  int ret;
  FILE* fp;

  if ((fp = fopen(filename, "r")) == NULL)
  {
    return FALSE;
  }

  //read the tone mapping config file
  for (idx = 0; idx < qmatrix_size; idx++)
  {
    ret = fscanf(fp, "%d", &dst_qmatrix[idx]);
    if (ret!=1)
    {
      return FALSE;
    }
  }
  return TRUE;
}

#endif //defined(NEC_MEMORY_COMPRESSION)
